"""SenseME Discovery.

This class periodically broadcasts discovery packets and listens for response messages
from SenseME devices by Big Ass Fans. Discovered devices are announced via callbacks.

Based on work from Bruce at http://bruce.pennypacker.org/tag/senseme-plugin/
and https://github.com/bpennypacker/SenseME-Indigo-Plugin

Based on work from TomFaulkner at https://github.com/TomFaulkner/SenseMe

Source can be found at https://github.com/mikelawrence/aiosenseme
"""
# from aiosenseme.aiosenseme.scripts.commandline import _DEVICES
import asyncio
import errno
import inspect
import ipaddress
import logging
import random
import socket
import sys
import time
import traceback

import ifaddr

from .device import (
    DEVICE_TYPES,
    IGNORE_MODELS,
    SensemeDevice,
    SensemeFan,
    SensemeLight,
)

_LOGGER = logging.getLogger(__name__)

PORT = 31415

# the default windows event loop (ProactorEventLoop) does not support
# create_datagram_endpoint() needed by this module. Switching to the
# SelectorEventLoop fixes this problem but may have unintended consequences.
if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    asyncio.get_event_loop()  # start the loop
    _LOGGER.debug("Changing event loop for Windows")


class SensemeDiscoveryEndpoint:
    """High-level endpoint for SenseME Discovery protocol."""

    # one receive queue for all endpoints
    receive_queue = asyncio.Queue()

    def __init__(self, ip: str = None):
        """Initialize Senseme Discovery Endpoint."""
        self.opened = False
        self.transport = None
        self.ip = ip

    def abort(self):
        """Close the transport immediately.

        Buffered write data will be lost.
        """
        self.opened = False
        if self.transport is None:
            return
        self.transport.abort()
        self.close()

    def close(self):
        """Close the transport gracefully.

        Buffered write data will be sent.
        """
        self.opened = False
        if self.transport is None:
            return
        self.receive_queue.put_nowait(None)  # tell receive() socket is closed
        if self.transport:
            self.transport.close()

    def is_closing(self) -> bool:
        """Return True if the endpoint is closed or closing."""
        if not self.opened:
            return True  # unopened connection is closed
        if self.transport is None:
            return True  # opened connection but no transport is closed
        return self.transport.is_closing()

    async def receive(self) -> SensemeDevice:
        """Wait for discovered device and return it.

        Return None when the socket is closed.
        This method is a coroutine.
        """
        while True:
            if self.receive_queue.empty() and self.transport.is_closing():
                return None
            rsp = await self.receive_queue.get()
            if rsp is None:
                return None
            msg = rsp[0]
            addr = rsp[1]
            if len(msg) > 200 or len(msg) < 31:
                continue
            if msg[0] != "(" or msg[-1] != ")":
                continue
            msg = msg[1:-1]
            msg_data = msg.split(";")
            if len(msg_data) != 5:
                continue
            if msg_data[4].upper() in IGNORE_MODELS:
                _LOGGER.debug("Ignored '%s' on %s", msg, self.ip)
                continue
            _LOGGER.debug("Received '%s' from %s on %s", msg, addr, self.ip)
            device_type = DEVICE_TYPES.get(msg_data[4], "FAN")
            if device_type == "FAN":
                device = SensemeFan(
                    name=msg_data[0],
                    mac=msg_data[3],
                    address=addr,
                    base_model=msg_data[4],
                )
            elif device_type == "LIGHT":
                device = SensemeLight(
                    name=msg_data[0],
                    mac=msg_data[3],
                    address=addr,
                    base_model=msg_data[4],
                )
            return device

    def send_broadcast(self):
        """Send the SenseME Discovery broadcast packet."""
        if not self.is_closing():
            data = "<ALL;DEVICE;ID;GET>".encode("utf-8")
            self.transport.sendto(data, ("<broadcast>", PORT))
            _LOGGER.debug("Discovery broadcast on %s", self.ip)


class SensemeDiscoveryProtocol(asyncio.DatagramProtocol):
    """Datagram protocol for SenseME Discovery."""

    def __init__(self, endpoint: SensemeDiscoveryEndpoint):
        """Initialize Senseme Discovery Protocol."""
        self._endpoint = endpoint

    # Protocol methods
    def connection_made(self, transport: asyncio.Protocol):
        """Socket connect on SenseME Discovery Protocol."""
        self._endpoint.transport = transport
        self._endpoint.opened = True
        _LOGGER.debug("Listening on %s", self._endpoint.ip)
        self._endpoint.send_broadcast()

    def connection_lost(self, exc):  # pylint: disable=unused-argument
        """Lost connection SenseME Discovery Protocol."""
        _LOGGER.debug("Listener closed on %s", self._endpoint.ip)
        self._endpoint.close()  # half-closed connections are not permitted

    # Datagram protocol methods
    def datagram_received(self, data: str, addr: str):
        """UDP packet received on SenseME Discovery Protocol."""
        if data:
            msg = data.decode("utf-8")
            try:
                self._endpoint.receive_queue.put_nowait((msg, addr[0]))
            except asyncio.QueueFull:
                _LOGGER.error("Receive queue full")

    def error_received(self, exc):
        """Error on SenseME Discovery Protocol."""
        _LOGGER.debug(
            "Protocol error on %s\n%s", self._endpoint.ip, traceback.format_exc()
        )
        self._endpoint.close()


class SensemeDiscovery:
    """SenseME Discovery Class.

    This class periodically broadcasts discovery packets and listens
    for response messages from SenseME devices by Big Ass Fans.
    """

    _devices = []  # all SensemeDiscovery objects use the same device list

    def __init__(self, start_first: bool = True, refresh_minutes: int = 5):
        """Initialize Senseme Discovery Protocol."""
        self.start_first = start_first
        self.refresh_minutes = refresh_minutes
        self._is_running = False
        self._callbacks = []
        self._broadcaster_task = None

    @property
    def devices(self):
        """Get the current list of discovered devices."""
        return self._devices

    async def async_add_by_device_info(self, info: str):
        """Add a device by IP address."""
        for device in self._devices:
            if device == info["address"]:
                _LOGGER.debug("Did not add by device info. Device already exists")
                return
        status, device = await async_get_device_by_device_info(
            info=info,
            start_first=self.start_first,
            refresh_minutes=self.refresh_minutes,
        )
        if device is None:
            _LOGGER.debug(
                "Add device by device info failed. Unable to connect '%s'",
                info.get("address"),
            )
            return
        if device not in self._devices:
            self._devices.append(device)
            _LOGGER.debug("Add by device info %s", device)
        else:
            _LOGGER.debug("Did not add by device info. Device already exists", device)
            return
        for callback in self._callbacks:
            if inspect.iscoroutinefunction(callback):
                asyncio.create_task(callback(self._devices.copy()))
            else:
                callback(self._devices.copy())

    async def async_add_by_ip_address(self, address: str):
        """Add a device by IP address."""
        for device in self._devices:
            if device == address:
                _LOGGER.debug("Did not add by IP address. Device already exists")
                return
        device = await async_get_device_by_ip_address(
            address=address, refresh_minutes=self.refresh_minutes, timeout_seconds=5
        )
        if device is None:
            _LOGGER.debug(
                "Add device by IP address failed. Unable to connect '%s'", address
            )
            return
        if self.start_first:
            if device not in self._devices:
                await device.async_update()
                if device not in self._devices:
                    self._devices.append(device)
                    _LOGGER.debug("Add by IP address %s", device)
                else:
                    _LOGGER.debug("Did not add by IP address. Device already exists")
                    return
        else:
            if device not in self._devices:
                self._devices.append(device)
                _LOGGER.debug("Add by IP address %s", device)
            else:
                _LOGGER.debug("Did not by IP address. Device already exists")
                return
        for callback in self._callbacks:
            if inspect.iscoroutinefunction(callback):
                asyncio.create_task(callback(self._devices.copy()))
            else:
                callback(self._devices.copy())

    def add_callback(self, callback):
        """Add callback function/coroutine.

        Called when parameters are updated.
        """
        if callback not in self._callbacks:
            self._callbacks.append(callback)
            if inspect.iscoroutinefunction(callback):
                asyncio.create_task(callback(self._devices.copy()))
                _LOGGER.debug("Added coroutine callback")
            else:
                callback(self._devices.copy())
                _LOGGER.debug("Added function callback")

    def remove_callback(self, callback):
        """Remove existing callback function/coroutine."""
        if callback in self._callbacks:
            self._callbacks.remove(callback)
            _LOGGER.debug("Removed callback")

    async def _create_endpoints(self):
        """Create an endpoint per interface and return as a list of endpoints."""
        endpoints = []
        loop = asyncio.get_running_loop()
        listening = 0
        for adapter in ifaddr.get_adapters():
            for ip in adapter.ips:
                if ip.is_IPv4 and not ipaddress.ip_address(ip.ip).is_loopback:
                    # _LOGGER.debug("Found IPv4 %s", ip.ip)
                    try:
                        endpoint = SensemeDiscoveryEndpoint(ip.ip)
                        await loop.create_datagram_endpoint(
                            lambda ep=endpoint: SensemeDiscoveryProtocol(ep),
                            local_addr=(ip.ip, PORT),
                            family=socket.AF_INET,
                            allow_broadcast=True,
                        )
                        endpoints.append(endpoint)
                        listening += 1
                    except OSError as e:
                        # borrowed this error handling from python-zeroconf
                        _errno = e.args[0]
                        err_einval = {errno.EINVAL}
                        if sys.platform == "win32":
                            err_einval |= {errno.WSAEINVAL}
                        if _errno == errno.EADDRINUSE:
                            _LOGGER.debug(
                                "Unable to listen on %s. " "Address already in use",
                                ip.ip,
                            )
                        elif _errno == errno.EADDRNOTAVAIL:
                            _LOGGER.debug(
                                "Unable to listen on %s. " "Address not available",
                                ip.ip,
                            )
                        elif _errno in err_einval:
                            _LOGGER.debug(
                                "Unable to listen on %s. " "Multicast not supported",
                                ip.ip,
                            )
                        else:
                            _LOGGER.error(
                                "Unable to listen on %s.\n%s",
                                traceback.format_exc(),
                                ip.ip,
                            )

        if listening == 0:
            # failed to bind to any address
            _LOGGER.error("Failed to listen on any address.")

        return endpoints

    async def _broadcaster(self):
        """Periodically broadcast discovery packet.

        If the underlying socket has an error this task will exit.
        This method is a coroutine.
        """
        self._is_running = True
        device = None
        while True:
            try:
                endpoints = await self._create_endpoints()
                start = time.time()
                while True:
                    try:
                        device = await asyncio.wait_for(endpoints[0].receive(), 1)
                    except asyncio.TimeoutError:
                        device = None
                        if time.time() - start < 5:
                            for endpoint in endpoints:
                                endpoint.send_broadcast()
                        else:
                            for endpoint in endpoints:
                                endpoint.abort()
                            endpoints = []
                            break
                    if device is not None:
                        if device not in self._devices:
                            if self.start_first:
                                if await device.async_update():
                                    if device not in self._devices:
                                        self._devices.append(device)
                                        _LOGGER.debug("Discovered %s", device)
                                else:
                                    _LOGGER.debug("Failed to start %s", device.name)
                            else:
                                if await device.async_fill_out_info():
                                    if device not in self._devices:
                                        self._devices.append(device)
                                        _LOGGER.debug("Discovered %s", device)
                                else:
                                    _LOGGER.debug(
                                        "Failed to retrieve secondary info for %s",
                                        device.name,
                                    )
                        for callback in self._callbacks:
                            if inspect.iscoroutinefunction(callback):
                                asyncio.create_task(callback(self._devices.copy()))
                            else:
                                callback(self._devices.copy())
                await asyncio.sleep(1)
                wait = self.refresh_minutes * 60 + random.uniform(-10, 10)
                _LOGGER.debug("Currently %s known senseme devices", len(self._devices))
                _LOGGER.debug("Discovery waiting for %s seconds", int(wait))
                await asyncio.sleep(wait)
            except asyncio.CancelledError:
                _LOGGER.debug("Broadcaster task cancelled")
                return
            except IndexError:
                # Caused by not listening on any endpoint.
                # _create_endpoints() already reported the error
                return
            except Exception:
                _LOGGER.error("Broadcaster task error\n%s", traceback.format_exc())
                raise
            finally:
                for endpoint in endpoints:
                    endpoint.abort()
                endpoints = None
        _LOGGER.error("Broadcaster task ended")

    def start(self):
        """Start broadcaster task.

        Will maintain a list of discovered devices.
        """
        if not self._is_running:
            self._broadcaster_task = asyncio.create_task(self._broadcaster())

    def stop(self):
        """Stop broadcaster task.

        Any discovered devices will remain in memory and will continue to update.
        """
        if self._is_running is True:
            self._broadcaster_task.cancel()
            self._is_running = False

    def remove_discovered_devices(self):
        """Stop broadcaster task.

        Any discovered devices will be stopped and removed from memory.
        """
        self.stop()
        for device in self._devices:
            device.stop()
        self._devices = []


async def discover_all(timeout_seconds=5) -> bool:
    """Returns all SenseME devices found on the network.

    This function will always take timeout_seconds to complete.
    This method is a coroutine.
    """
    start = time.time()
    discovery = SensemeDiscovery(False)
    try:
        discovery.start()
        while True:
            await asyncio.sleep(0.1)
            for device in discovery.devices:
                await device.async_fill_out_info()
            if time.time() - start > timeout_seconds:
                return discovery.devices.copy()
    finally:
        count = len(discovery.devices)
        _LOGGER.debug("Discovered %s device%s", count, "" if count == 1 else "s")
        discovery.stop()


async def discover_any(timeout_seconds=5) -> bool:
    """Return True if any SenseME devices are found on the network.

    This function will always take timeout_seconds to complete.
    This method is a coroutine.
    """
    discovery = SensemeDiscovery(True, 1)
    discovery.start()
    await asyncio.sleep(timeout_seconds)
    count = len(discovery.devices)
    discovery.stop()
    _LOGGER.debug("Discovered %s device%s", count, "" if count == 1 else "s")
    return count > 0


async def discover(value, timeout_seconds=5) -> SensemeDevice:
    """Discover a device with a name or room name.

    None is returned if the device was not found.
    This function will take up timeout_seconds to complete.
    This method is a coroutine.
    """
    start = time.time()
    discovery = SensemeDiscovery(False)
    try:
        discovery.start()
        while True:
            await asyncio.sleep(0.1)
            devices = discovery.devices.copy()
            for device in devices:
                if device == value:
                    await device.async_fill_out_info()
                    device.start()
                    await device.async_update()
                    return device
            if time.time() - start > timeout_seconds:
                return None
    finally:
        discovery.stop()


# pylint: disable=protected-access
async def async_get_device_by_ip_address(
    address: str, refresh_minutes: int = 1, timeout_seconds: int = 5
) -> SensemeDevice:
    """Asynchronously get a device by IP Address.

    This function will connect to the device to determine the model so an
    appropriate SensemeFan or SensmeLight object can be returned.
    Note the device will not be started.
    If connection to the device was unsuccessful None will be returned.
    This function will take up timeout_seconds to complete or fail.
    """
    try:
        # do some checking on IP address
        ipaddress.ip_address(address)
    except ValueError:
        _LOGGER.debug(
            "Get device by IP address (%s) failed. Invalid IP address", address
        )
        return None
    try:
        basedevice = SensemeDevice(address=address)
        if await asyncio.wait_for(basedevice.async_fill_out_info(), timeout_seconds):
            if basedevice.device_type == "FAN":
                device = SensemeFan(
                    name=basedevice._name,
                    mac=basedevice._mac,
                    address=address,
                    base_model=basedevice._base_model,
                    refresh_minutes=refresh_minutes,
                )
            else:
                device = SensemeLight(
                    name=basedevice._name,
                    mac=basedevice._mac,
                    address=address,
                    base_model=basedevice._base_model,
                    refresh_minutes=refresh_minutes,
                )
            device._uuid = basedevice._uuid
            device._room_name = basedevice._room_name
            device._room_type = basedevice._room_type
            device._has_light = basedevice._has_light
            device._has_sensor = basedevice._has_sensor
            device._fw_name = basedevice._fw_name
            device._fw_version = basedevice._fw_version
            device._data = basedevice._data
            device._first_update = basedevice._first_update
            return device
    except asyncio.TimeoutError:
        _LOGGER.debug("Get device by IP address (%s) failed. No response", address)
    except asyncio.CancelledError:
        _LOGGER.debug("Task get device by IP address (%s) cancelled", address)
    return None


async def async_get_device_by_device_info(
    info: dict, start_first=False, refresh_minutes: int = 1, timeout_seconds: int = 10
) -> SensemeDevice:
    """Asynchronously get a device by device info dict.

    The device info dict comes from SensemeDevice.get_device_info(). This
    method will use the device info dict to determine the model so an
    appropriate SensemeFan or SensmeLight object can be returned. This
    method will also either start the device or retrieve secondary
    information based on start_first.
    A Tuple of status and the device are returned. When status is True the
    device was either started or secondary information was retrieved from
    the device based on start_first. False means failed to connect to device.
    This method will take up timeout_seconds to complete or fail.
    """
    try:
        if info.get("is_fan"):
            device = SensemeFan(info=info, refresh_minutes=refresh_minutes)
        elif info.get("is_light"):
            device = SensemeLight(info=info, refresh_minutes=refresh_minutes)
        if start_first:
            if await device.async_update(timeout_seconds=timeout_seconds):
                return [True, device]
        else:
            if await asyncio.wait_for(device.async_fill_out_info(), timeout_seconds):
                return [True, device]
    except asyncio.TimeoutError:
        _LOGGER.debug("Get device by device info dict (%s) failed. No response", info)
    except asyncio.CancelledError:
        _LOGGER.debug("Task get device by device info dict(%s) cancelled", info)
    return [False, device]



"""Script interface for aiosenseme library."""

import argparse
import asyncio
import json
import logging
from typing import List

import aiosenseme
from aiosenseme import SensemeDevice, SensemeDiscovery, __version__

ARGS = argparse.ArgumentParser(
    description="Discover and control SenseME devices by Big Ass Fans."
)
ARGS.add_argument(
    "-V",
    "--version",
    action="store_true",
    dest="version",
    default=False,
    help="display version number",
)
ARGS.add_argument(
    "--debug",
    action="store_true",
    dest="debug",
    default=False,
    help="enable debug level logging",
)
ARGS.add_argument(
    "--listen",
    action="store_true",
    dest="listen",
    default=False,
    help="Connect to device and show changes to the fan",
)
ARGS.add_argument(
    "-j",
    "--json",
    action="store_true",
    dest="json",
    default=False,
    help="return device information and state as json",
)
ARGS.add_argument(
    "-d",
    "--discover",
    action="store_true",
    dest="discover",
    default=False,
    help="discover all SenseME devices on the network",
)
ARGS.add_argument(
    "-m",
    "--models",
    action="store_true",
    dest="models",
    default=False,
    help="list known SenseME device models",
)
ARGS.add_argument(
    "-i",
    "--ip",
    action="store",
    dest="ip",
    default=None,
    help="IP Address",
)
ARGS.add_argument(
    "-n",
    "--name",
    action="store",
    dest="name",
    default=None,
    help="SenseME device name",
)
ARGS.add_argument(
    "-f",
    "--fan",
    action="store",
    dest="fan",
    default=None,
    choices=["on", "off"],
    help="fan power",
)
ARGS.add_argument(
    "-s",
    "--speed",
    action="store",
    dest="speed",
    default=None,
    type=int,
    choices=range(1, 8),
    help="fan speed",
)
ARGS.add_argument(
    "-l",
    "--light",
    action="store",
    dest="light",
    default=None,
    choices=["on", "off"],
    help="light power",
)
ARGS.add_argument(
    "-b",
    "--brightness",
    action="store",
    dest="brightness",
    default=None,
    type=int,
    choices=range(0, 17),
    help="light brightness",
)
ARGS.add_argument(
    "-c",
    "--colortemp",
    action="store",
    dest="colortemp",
    default=None,
    type=int,
    choices=range(2200, 5100, 100),
    help="light color temperature",
)
ARGS.add_argument(
    "-w",
    "--whoosh",
    action="store",
    dest="whoosh",
    default=None,
    choices=["on", "off"],
    help="fan whoosh mode",
)

# array of discovered devices
_DEVICES = []


def print_device(device: SensemeDevice):
    """Print information about a device."""
    msg = f"{device.name}\n"
    if device.room_status:
        msg += f"  Room Name: {device.room_name}, Room Type: {device.room_type}\n"
    if device.is_fan:
        if device.has_light:
            msg += f"  Model: {device.model} with light, "
        else:
            msg += f"  Model: {device.model} without light, "
    elif device.is_light:
        msg += f"  Model: {device.model}, "
    else:
        msg += f"  Model: {device.model}, "
    msg += f"FW Version: {device.fw_version}\n"
    msg += f"  IP Addr: {device.address}, MAC Addr: {device.mac}\n"
    msg += f"  UUID: {device.uuid}"
    print(msg)


def print_state(prefix: str, device: SensemeDevice):
    """Print information about a devices's current state."""
    msg = prefix
    if device.is_fan:
        if device.fan_on:
            msg += f": Fan is on (speed: {device.fan_speed}"
            if device.fan_whoosh_mode:
                msg += ", whoosh mode is on"
            msg += ")"
        else:
            msg += ": Fan is off"
        if device.light_on:
            msg += f", Light is on (brightness: {device.light_brightness})"
        else:
            msg += ", Light is off"
        if device.sleep_mode:
            msg += ", Sleep Mode is on"
    elif device.is_light:
        if device.light_on:
            msg += f": Light is on (brightness: {device.light_brightness}"
            msg += f", color temp: {device.light_color_temp})"
        else:
            msg += ": Light is off"
    else:
        msg += ": Unknown SenseME device"
    print(msg)


async def discovered(devices: List[SensemeDevice]):
    """Discovered SenseME device callback function.

    Called when discovery has detected a SenseME device.
    Each time a device is discovered all devices discovered are reported.
    """
    global _DEVICES
    for device in devices:
        if device not in _DEVICES:
            _DEVICES.append(device)
            print_device(device)


async def process_args():
    """Process command line arguments."""
    try:
        device = None
        args = ARGS.parse_args()
        if args.debug or args.listen:
            logging.basicConfig(level=logging.DEBUG)
        if args.version is True:
            print("aiosenseme %s" % __version__)
            print("Copyright (C) 2021 by Mike Lawrence")
            print(
                "This is free software. "
                "You may redistribute copies of it under the terms"
            )
            print(
                "of the GNU General Public License "
                "<http://www.gnu.org/licenses/gpl.html>."
            )
            print("There is NO WARRANTY, to the extent permitted by law.")
            return
        if args.discover is True:
            try:
                print("Attempting to discover SenseME devices...")
                discovery = SensemeDiscovery(start_first=True)
                discovery.add_callback(discovered)
                discovery.start()
                await asyncio.sleep(5)
                count = len(discovery.devices)
                if count == 0:
                    print("Found no SenseME devices.")
                else:
                    if count == 1:
                        plural = ""
                    else:
                        plural = "s"
                    print(f"Found {count} SenseME device{plural}.")
            finally:
                discovery.stop()
                return
        if args.models is True:
            msg = "Known SenseME models: "
            first = True
            for model in SensemeDevice.models():
                if first:
                    first = False
                    msg += model
                else:
                    msg += ", " + model
            print(msg)
            return
        if not args.name and not args.ip:
            print("You must specify a SenseME device by using -n/--name or -i/--ip")
            return
        if args.ip is not None:
            device = await aiosenseme.async_get_device_by_ip_address(
                args.ip, timeout_seconds=5
            )
            if device is None:
                print(f"Device at IP '{args.ip}' not found")
                return
            await device.async_update(timeout_seconds=5)
        elif args.name is not None:
            device = await aiosenseme.discover(args.name, 5)
            if device is None:
                print(f"Name or Room '{args.name}' not found")
                return
        if args.listen:
            while True:
                await asyncio.sleep(1.0)
        if args.json:
            info = device.get_device_info
            info["model"] = device.model
            info["fw_version"] = device.fw_version
            info["device_indicators"] = device.device_indicators
            info["device_beeper"] = device.device_beeper
            info["network_ip"] = device.network_ip
            info["network_subnetmask"] = device.network_subnetmask
            info["network_gateway"] = device.network_gateway
            info["network_ssid"] = device.network_ssid
            info["room_name"] = device.room_name
            info["room_type"] = device.room_type
            info["motion_detected"] = device.motion_detected
            info["sleep_mode"] = device.sleep_mode
            if device.is_fan:
                info["fan_on"] = device.fan_on
                info["fan_speed"] = device.fan_speed
                info["fan_speed_min"] = device.fan_speed_min
                info["fan_speed_max"] = device.fan_speed_max
                info["fan_speed_limits_room"] = device.fan_speed_limits_room
                info["fan_dir"] = device.fan_dir
                info["fan_whoosh_mode"] = device.fan_whoosh_mode
                info["fan_autocomfort"] = device.fan_autocomfort
                info["fan_smartmode"] = device.fan_smartmode
                info["fan_cooltemp"] = device.fan_cooltemp
                info["motion_fan_auto"] = device.motion_fan_auto
            if device.is_light:
                info["light_color_temp"] = device.light_color_temp
                info["light_color_temp_min"] = device.light_color_temp_min
                info["light_color_temp_max"] = device.light_color_temp_max
            if device.has_light:
                info["light_on"] = device.light_on
                info["light_brightness"] = device.light_brightness
                info["light_brightness_min"] = device.light_brightness_min
                info["light_brightness_max"] = device.light_brightness_max
                info[
                    "light_brightness_limits_room"
                ] = device.light_brightness_limits_room
                info["motion_light_auto"] = device.motion_light_auto
            print(json.dumps(info, sort_keys=True, indent=4))
            return
        print_device(device)
        print_state("State", device)
        changed = False
        if device.is_fan:
            if args.whoosh is not None:
                print(f"whoosh={args.whoosh}")
                if device.fan_whoosh_mode != (args.whoosh == "on"):
                    changed = True
                device.fan_whoosh_mode = args.whoosh == "on"
            if args.speed is not None:
                if args.fan is not None:
                    print(
                        "When specifying --fanspeed there is no " "reason to set --fan"
                    )
                if device.fan_speed != args.speed:
                    changed = True
                device.fan_speed = args.speed
            elif args.fan is not None:
                if device.fan_on != (args.fan == "on"):
                    changed = True
                device.fan_on = args.fan == "on"
            if device.has_light:
                if args.colortemp is not None:
                    print("Fan lights do not have adjustable color temperature")
                if args.brightness is not None:
                    if args.light is not None:
                        print(
                            "When specifying --brightness there is no "
                            "reason to set --light"
                        )
                    if device.light_brightness != args.brightness:
                        changed = True
                    device.light_brightness = args.brightness
                elif args.light is not None:
                    if device.light_on != (args.light == "on"):
                        changed = True
                    device.light_on = args.light == "on"
            else:
                if (
                    args.brightness is not None
                    or args.light is not None
                    or args.colortemp is not None
                ):
                    print("Fan does not have a light to adjust")
        else:
            if args.brightness is not None:
                if args.light is not None:
                    print(
                        "When specifying --brightness there is no "
                        "reason to set --light"
                    )
                if device.light_brightness != args.brightness:
                    changed = True
                device.light_brightness = args.brightness
            elif args.light is not None:
                if device.light_on != (args.light == "on"):
                    changed = True
                device.light_on = args.light == "on"
            if args.colortemp is not None:
                if device.light_color_temp != args.colortemp:
                    changed = True
                device.light_color_temp = args.colortemp
        if changed:
            await asyncio.sleep(1.0)
            print_state("New State", device)
    finally:
        if device is not None:
            device.stop()


def cli():
    """Command line interface for SensemeDiscovery."""
    task = asyncio.Task(process_args())
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    cli()



"""SensemeFan Class.

This class connects to a SenseME fan by Big Ass Fans to retrieve and maintain its
complete state. It keep a connection always open to the fan and will receive virtually
instantaneous state changes (push) from any source (i.e. remote control or Haiku App).
The always open connection will automatically reconnect on errors or when the SenseME
fan disconnects. Control and access to the fan's current state is provided by class
properties. State changes are announced via callbacks.

Based on work from Bruce at http://bruce.pennypacker.org/tag/senseme-plugin/
and https://github.com/bpennypacker/SenseME-Indigo-Plugin

Based on work from TomFaulkner at https://github.com/TomFaulkner/SenseMe

Source can be found at https://github.com/mikelawrence/aiosenseme
"""
import asyncio
import inspect
import ipaddress
import logging
import random
import time
import traceback
from typing import Any, Callable, Tuple

_LOGGER = logging.getLogger(__name__)

PORT = 31415

ONOFF = ["ON", "OFF"]
DIRECTIONS = ["FWD", "REV"]
AUTOCOMFORTS = ["OFF", "COOLING", "HEATING", "FOLLOWTSTAT"]
INVALID_DATA = "?????"  # data we never expect to be sent by the device
ROOM_TYPES = [
    "Undefined",  # 0, not in a room
    "Other",  # 1
    "Master Bedroom",  # 2
    "Bedroom",  # 3
    "Den",  # 4
    "Family Room",  # 5
    "Living Room",  # 6
    "Kids Room",  # 7
    "Kitchen",  # 8
    "Dining Room",  # 9
    "Basement",  # 10
    "Office",  # 11
    "Patio",  # 12
    "Porch",  # 13
    "Hallway",  # 14
    "Entryway",  # 15
    "Bathroom",  # 16
    "Laundry",  # 17
    "Stairs",  # 18
    "Closet",  # 19
    "Sunroom",  # 20
    "Media Room",  # 21
    "Gym",  # 22
    "Garage",  # 23
    "Outside",  # 24
    "Loft",  # 25
    "Playroom",  # 26
    "Pantry",  # 27
    "Mudroom",  # 28
]

DEVICE_MODELS = {
    "FAN,HAIKU,SENSEME": "Haiku Fan",
    "FAN,HAIKU,HSERIES": "Haiku Fan",  # H Series is now called plain Haiku
    "FAN,LSERIES": "Haiku L Fan",
    "LIGHT,HAIKU": "Haiku Light",
}

DEVICE_TYPES = {
    "FAN,HAIKU,SENSEME": "FAN",
    "FAN,HAIKU,HSERIES": "FAN",
    "FAN,LSERIES": "FAN",
    "LIGHT,HAIKU": "LIGHT",
}

IGNORE_MODELS = [
    "SWITCH,SENSEME",
]

SUPPRESS_CALLBACK_PARAMS = {"SLEEP;EVENT"}


class SensemeEndpoint:
    """High-level endpoint for SenseME protocol."""

    def __init__(self):
        """Initialize Senseme Discovery Endpoint."""
        self.receive_queue = asyncio.Queue()
        self.opened = False
        self.transport = None

    def abort(self):
        """Close the transport immediately. Buffered write data will be flushed."""
        self.transport.abort()
        self.close()

    def close(self):
        """Close the transport gracefully. Buffered write data will be sent."""
        if self.transport is None:
            return
        self.receive_queue.put_nowait(None)  # tell receive() socket is closed
        if self.transport:
            self.transport.close()

    def is_closing(self) -> bool:
        """Return True if the endpoint is closed or closing."""
        if not self.opened:
            return False  # unopened connection is not closed
        if self.transport is None:
            return True  # opened connection but no transport is closed
        return self.transport.is_closing()

    async def receive(self) -> str:
        """Wait for a message from the SenseME fan.

        Return None when the socket is closed.
        This method is a coroutine.
        """
        if self.receive_queue.empty() and self.transport.is_closing():
            return None
        return await self.receive_queue.get()

    def send(self, cmd):
        """Send a command to the SenseME fan."""
        self.transport.write(cmd.encode("utf-8"))


class SensemeProtocol(asyncio.Protocol):
    """Protocol for SenseME communication."""

    def __init__(self, name, endpoint: SensemeEndpoint):
        """Initialize Senseme Protocol."""
        self._name = name
        self._endpoint = endpoint

    # Protocol methods
    def connection_made(self, transport: asyncio.Protocol):
        """Socket connect on SenseME Protocol."""
        _LOGGER.debug("%s: Connected", self._name)
        self._endpoint.transport = transport
        self._endpoint.opened = True

    def connection_lost(self, exc):  # pylint: disable=unused-argument
        """Lost connection SenseME Protocol."""
        _LOGGER.debug("%s: Connection lost", self._name)
        self._endpoint.close()  # half-closed connections are not permitted

    # Streaming Protocol methods
    def data_received(self, data: str) -> str:
        """UDP packet received on SenseME Protocol."""
        if data:
            msg = data.decode("utf-8")
            try:
                self._endpoint.receive_queue.put_nowait(msg)
            except asyncio.QueueFull:
                _LOGGER.error("%s: Receive queue full", self._name)

    def eof_received(self) -> bool:
        """EOF received on SenseME Protocol."""
        return False  # tell the transport to close itself


class SensemeDevice:
    """SensemeDevice base class."""

    def __init__(
        self,
        info: dict = None,
        name: str = None,
        uuid: str = None,
        mac: str = None,
        address: str = None,
        base_model: str = None,
        refresh_minutes: int = 1,
    ):
        """Initialize SensemeDevice Class."""
        self.refresh_minutes = refresh_minutes
        if info is not None:
            self._name = info.get("name", None)
            self._uuid = info.get("uuid", None)
            self._mac = info.get("mac", None)
            self._address = info.get("address", None)
            self._base_model = info.get("base_model", None)
            self._has_light = info.get("has_light", None)
            self._has_sensor = info.get("has_sensor", None)
        else:
            self._name = name
            self._uuid = uuid
            self._mac = mac
            self._address = address
            self._base_model = base_model
            self._has_light = None
            if base_model is not None:
                if self.model in ["Haiku Fan", "Haiku Light"]:
                    self._has_sensor = True
                elif self.model == "Haiku L Fan":
                    self._has_sensor = None
                else:
                    self._has_sensor = False
            else:
                self._has_sensor = None
        self._room_name = None
        self._room_type = None
        self._fw_name = "Unknown"
        self._fw_version = None

        self._data = dict()
        self._is_running = False
        self._is_connected = False
        self._connection_lost = False
        self._endpoint = None
        self._listener_task = None
        self._updater_task = None
        self._error_count = 0
        self._leftover = ""
        self._callbacks = []
        self._coroutine_callbacks = []
        self._first_update = False

    def __eq__(self, other: Any) -> bool:
        """Equals magic method."""
        try:
            ip_addr = ipaddress.ip_address(other)
            return ip_addr == ipaddress.ip_address(self._address)
        except ValueError:
            pass
        if isinstance(other, SensemeDevice):
            if self._address and other.address:
                return self._address == other.address
        if isinstance(other, str):
            return other in [self._name]

        return NotImplemented

    def __hash__(self) -> int:
        """Hash magic method."""
        return hash(self._uuid)

    @property
    def is_sec_info_complete(self) -> bool:
        """Return if all secondary information is complete."""
        if self._name is None:
            return False
        if self._uuid is None:
            return False
        if self._mac is None:
            return False
        if self._base_model is None:
            return False
        if self._room_name is None:
            return False
        if self._room_type is None:
            return False
        if self._has_light is None:
            return False
        if self._has_sensor is None:
            return False
        if self._fw_version is None:
            return False
        if self._base_model is None:
            return False
        return True

    @property
    def get_device_info(self) -> dict:
        """Get a dict with all key information abouth this device."""
        return {
            "name": self._name,
            "uuid": self._uuid,
            "mac": self._mac,
            "address": self._address,
            "base_model": self._base_model,
            "has_light": self._has_light,
            "has_sensor": self._has_sensor,
            "is_fan": self.is_fan,
            "is_light": self.is_light,
        }

    @property
    def is_fan(self) -> str:
        """Return True if this device is a fan."""
        return False

    @property
    def is_light(self) -> str:
        """Return True if the device is a standalone light."""
        return False

    @property
    def name(self) -> str:
        """Return name of device."""
        if self._name is None:
            if self._mac is None:
                return self._address
            return self._mac
        return self._name

    @property
    def uuid(self) -> str:
        """Return UUID for this device.

        This is the network token read from the device.
        It will be unique for each device.
        """
        return self._uuid

    @property
    def mac(self) -> str:
        """Return MAC address of device."""
        return self._mac

    @property
    def address(self) -> str:
        """Return IP address of device."""
        return self._address

    @property
    def available(self) -> bool:
        """Return True when device is connected and all parameters have been updated."""
        return self._is_connected and self._first_update

    @property
    def connected(self) -> bool:
        """Return True when device is connected."""
        return self._is_connected

    @property
    def device_type(self) -> str:
        """Return type of device.

        If the model is unknown then the default response is a "FAN".
        """
        return DEVICE_TYPES.get(self._base_model.upper(), "FAN")

    @property
    def model(self) -> str:
        """Return Model of device."""
        return DEVICE_MODELS.get(self._base_model.upper(), self._base_model.upper())

    @property
    def base_model(self) -> str:
        """Return Model of device as reported."""
        return self._base_model

    @classmethod
    def models(cls) -> list:
        """Return list of possible model names."""
        no_duplicates = []
        for model_name in DEVICE_MODELS.values():
            if model_name not in no_duplicates:
                no_duplicates.append(model_name)
        return no_duplicates

    @property
    def is_unknown_model(self) -> bool:
        """Return True if the model is unknown."""
        return DEVICE_MODELS.get(self._base_model.upper(), None) is None

    @property
    def fw_version(self) -> str:
        """Return the version of the firmware running on the SenseME device."""
        return self._fw_version

    @property
    def has_light(self) -> bool:
        """Return True if the device has an installed light."""
        return self._has_light

    @property
    def has_sensor(self) -> bool:
        """Return True if the device has an occupancy sensor."""
        return self._has_sensor

    @property
    def device_indicators(self) -> str:
        """Return True if the device LED indicator is enabled."""
        value = self._data.get("DEVICE;INDICATORS", None)
        if value:
            return value == "ON"
        return None

    @device_indicators.setter
    def device_indicators(self, value: bool):
        """Enable/disable the device LED indicator."""
        if value:
            state = "ON"
        else:
            state = "OFF"
        self._send_command(f"DEVICE;INDICATORS;{state}")

    @property
    def device_beeper(self) -> bool:
        """Return the device audible alert enabled state."""
        status = self._data.get("DEVICE;BEEPER", None)
        if status:
            return status == "ON"
        return None

    @device_beeper.setter
    def device_beeper(self, value: bool):
        """Enable/disable the device audible alert."""
        if value:
            state = "ON"
        else:
            state = "OFF"
        self._send_command(f"DEVICE;BEEPER;{state}")

    @property
    def network_ap_on(self) -> bool:
        """Return the wireless access point running state."""
        status = self._data.get("NW;AP;STATUS", None)
        if status:
            return status == "ON"
        return None

    @property
    def network_dhcp_on(self) -> bool:
        """Return the device local DHCP service running state."""
        dhcp = self._data.get("NW;DHCP", None)
        if dhcp:
            return dhcp == "ON"
        return None

    @property
    def network_ip(self) -> str:
        """Return the network IP address of the device.

        This IP address is reported by the SenseME device and not necessarily the same IP
        address used to connect with the device.
        """
        addresses = self._data.get("NW;PARAMS;ACTUAL", None)
        if addresses:
            addresses = addresses.split(";")
            return addresses[0]
        return None

    @property
    def network_subnetmask(self) -> str:
        """Return the network gateway address of the device."""
        addresses = self._data.get("NW;PARAMS;ACTUAL", None)
        if addresses:
            addresses = addresses.split(";")
            return addresses[2]
        return None

    @property
    def network_gateway(self) -> str:
        """Return the network gateway address of the device."""
        addresses = self._data.get("NW;PARAMS;ACTUAL", None)
        if addresses:
            addresses = addresses.split(";")
            return addresses[1]
        return None

    @property
    def network_ssid(self) -> str:
        """Return the wireless SSID the device is connected to."""
        return self._data.get("NW;SSID", None)

    @property
    def network_token(self) -> str:
        """Return the network token of the device. This is the same as uuid."""
        return self._uuid

    @property
    def room_status(self) -> bool:
        """Return True if the device is in a room."""
        if self._room_name and self._room_type:
            return self._room_name != "EMPTY" and self._room_type != "0"
        return None

    @property
    def room_name(self) -> str:
        """Return the room name of the device.

        'EMPTY' is returned if not in a group.
        """
        return self._room_name

    @property
    def room_type(self) -> str:
        """Return the room type of the device."""
        room_type = self._room_type
        if room_type:
            if room_type >= len(ROOM_TYPES):
                room_type = 0
            return ROOM_TYPES[room_type]
        return None

    @property
    def light_on(self) -> bool:
        """Return True when light is on at any brightness."""
        state = self._data.get("LIGHT;PWR", None)
        if state:
            return state == "ON"
        return None

    @light_on.setter
    def light_on(self, state: bool):
        """Set the light power state."""
        value = "ON" if state else "OFF"
        self._send_command(f"LIGHT;PWR;{value}")

    @property
    def light_brightness(self) -> int:
        """Return the light brightness."""
        level = self._data.get("LIGHT;LEVEL;ACTUAL", None)
        if level:
            return int(level)
        return None

    @light_brightness.setter
    def light_brightness(self, level: int):
        """Set the light brightness."""
        if level < 0:
            level = 0
        if level > 16:
            level = 16
        self._send_command(f"LIGHT;LEVEL;SET;{level}")

    @property
    def light_brightness_min(self) -> int:
        """Return the light brightness minimum."""
        min_brightness = self._data.get("LIGHT;LEVEL;MIN", None)
        if min_brightness:
            return int(min_brightness)
        return None

    @property
    def light_brightness_max(self) -> int:
        """Return the light brightness maximum."""
        max_brightness = self._data.get("LIGHT;LEVEL;MAX", None)
        if max_brightness:
            return int(max_brightness)
        return None

    @property
    def light_brightness_limits_room(self) -> Tuple:
        """Return a tuple of the min and max light brightness for the room.

        A room can limit the minimum/maximum light brightness while keeping the same
        number of light brightness levels. On the Haiku by BAF application this setting
        can be found by clicking the room info button. You have to have at least one
        fan with installed light added to a room.
        """
        raw = self._data.get("LIGHT;BOOKENDS", None)
        if raw is None:
            return None
        values = raw.split(";")
        if len(values) != 2:
            return None
        min_bright = int(values[0])
        max_bright = int(values[1])
        return min_bright, max_bright

    @property
    def motion_detected(self) -> bool:
        """Return True when device motion sensor says room is occupied.

        Available on all SenseME fans.
        """
        status = self._data.get("SNSROCC;STATUS", None)
        if status:
            return status == "OCCUPIED"
        return None

    @property
    def motion_light_auto(self) -> bool:
        """Return True when light is in automatic on with motion mode."""
        if not self.has_light:
            return None
        state = self._data.get("LIGHT;AUTO", None)
        if state:
            return state == "ON"
        return None

    @motion_light_auto.setter
    def motion_light_auto(self, state: bool):
        """Set the light automatic on with motion mode."""
        if not self.has_light:
            return
        state = "ON" if state else "OFF"
        self._send_command(f"LIGHT;AUTO;{state}")

    @property
    def sleep_mode(self) -> bool:
        """Return True when sleep mode is enabled."""
        state = self._data.get("SLEEP;STATE", None)
        if state:
            return state == "ON"
        return None

    @sleep_mode.setter
    def sleep_mode(self, state: bool):
        """Set the sleep mode."""
        state = "ON" if state else "OFF"
        self._send_command(f"SLEEP;STATE;{state}")

    async def async_fill_out_info(self) -> bool:
        """Retrieve info from the SenseME device directly.

        Does not start background tasks to retrieve device/parameter information.
        This method is a coroutine.
        """
        writer = None
        if self.is_sec_info_complete:
            return True
        try:
            _LOGGER.debug(
                "Retrieve device information: Connecting to address %s", self.address
            )
            reader, writer = await asyncio.open_connection(self._address, PORT)
            _LOGGER.debug(
                "Retrieve device information: Status Update from address %s",
                self.address,
            )
            writer.write(f"<{self._address};DEVICE;ID;GET>".encode("utf-8"))
            writer.write(f"<{self._address};SNSROCC;STATUS;GET>".encode("utf-8"))
            writer.write(f"<{self._address};GETALL;GET>".encode("utf-8"))

            leftover = ""
            while True:
                # socket will throw a timeout error and abort this function if
                # no proper response is received
                line = await asyncio.wait_for(reader.readuntil(b")"), 10)
                leftover, _ = self._process_message(leftover + line.decode("utf-8"))
                if self._first_update:
                    return True
        except asyncio.TimeoutError:
            _LOGGER.debug(
                "Retrieve device information: Failed to connect to address %s",
                self.address,
            )
            return False
        except OSError:
            _LOGGER.debug(
                "%s: Retrieve device information: Error\n%s",
                self.address,
                traceback.format_exc(),
            )
            return False
        finally:
            # close the socket
            if writer is not None:
                _LOGGER.debug(
                    "Retrieve device information: Disconnecting from address %s",
                    self.address,
                )
                writer.close()
                await writer.wait_closed()

    def add_callback(self, callback: Callable):
        """Add callback function/coroutine. Called when parameters are updated."""
        is_coroutine = inspect.iscoroutinefunction(callback)
        if is_coroutine:
            if callback not in self._coroutine_callbacks:
                self._coroutine_callbacks.append(callback)
            _LOGGER.debug("%s: Added coroutine callback", self.name)
            return

        if callback not in self._callbacks:
            self._callbacks.append(callback)
        _LOGGER.debug("%s: Added function callback", self.name)

    def remove_callback(self, callback):
        """Remove existing callback function/coroutine."""
        if callback in self._coroutine_callbacks:
            self._coroutine_callbacks.remove(callback)
            _LOGGER.debug("%s: Removed coroutine callback", self.name)
            return
        if callback in self._callbacks:
            self._callbacks.remove(callback)
            _LOGGER.debug("%s: Removed function callback", self.name)
            return

    async def async_update(self, connection_lost=False, timeout_seconds=10) -> bool:
        """Wait for first update of all parameters in SenseME device.

        If connection_lost is True this connect will be treated as a reconnect.
        If timeout_seconds is how long this method will wait for a response
        before timing out. The Device will be started if not already.
        This method is a coroutine.
        """
        if connection_lost:
            self._connection_lost = True
        if not self._is_running:
            self.start()
        start = int(time.time())
        while not self.available:
            await asyncio.sleep(0.1)
            if int(time.time()) - start >= timeout_seconds:
                return False
        return True

    def _execute_callbacks(self):
        """Run all callbacks to indicate something has changed."""
        count = 0
        for callback in self._callbacks:
            count += 1
            callback()
        for callback in self._coroutine_callbacks:
            count += 1
            asyncio.create_task(callback())
        _LOGGER.debug("%s: %s callback(s) happened.", self.name, count)

    def _send_command(self, cmd):
        """Send a command to SenseME device."""
        msg = f"<{self.mac};{cmd}>"
        self._endpoint.send(msg)
        # _LOGGER.debug("%s: Command sent '%s'", self.name, cmd)

    def _process_message(self, line) -> str:
        """Process messages from device.

        May contain multiple responses in one line string.
        Last response may be a partial so it is returned.
        This partial should be added to next line.
        """
        should_callback = False
        # The line received may have multiple parenthesized responses
        # Split them and process each individual message.
        # Convert "(msg1)(msg2)(msg3)" to "(msg1)|(msg2)|(msg3)"
        # then split on the '|'
        for msg in line.replace(")(", ")|(").split("|"):
            if msg[-1] != ")":
                return msg
            # remove begining '(' and ending ')' from string
            # also extract name if undefined
            name, result = msg[1:-1].split(";", 1)
            if self._name is None:
                self._data["NAME;VALUE"] = name
                self._name = name
            # most messages have only one value at the end
            valuecount = 1
            if "BOOKENDS" in result:
                valuecount = 2
            elif "NW;PARAMS;ACTUAL" in result:
                valuecount = 3
            elif "DEVICE;LIGHT" in result:
                valuecount = len(result.split(";")) - 2
            elif "DEVICE;ID" in result:
                valuecount = 2
            # split on ';' and the associate the correct number of values
            values = result.split(";")
            key = ";".join(values[:-valuecount])
            value = ";".join(values[-valuecount:])
            if key == "ERROR":
                _LOGGER.error(
                    "%s: Command error response",
                    self.name,
                )
                continue
            if key == "TIME;VALUE":
                # ignore time parameter
                continue
            if self._data.get(key, INVALID_DATA) == value:
                # parameter has not changed, nothing to do
                continue
            self._data[key] = value  # update new key/value or changed value
            _LOGGER.debug("%s: Param updated: [%s]='%s'", self.name, key, value)
            if self.is_fan:
                if key == "WINTERMODE;STATE":
                    if not self._first_update:
                        self._first_update = True
                        _LOGGER.debug("%s: First Update Complete", self.name)
            else:
                if key == "SNSROCC;TIMEOUT;MIN":
                    if not self._first_update:
                        self._first_update = True
                        _LOGGER.debug("%s: First Update Complete", self.name)
            if self._first_update and key not in SUPPRESS_CALLBACK_PARAMS:
                should_callback = True
            # update certain local variables that are not part of data
            if key == "FW;NAME":
                self._fw_name = value
            elif key == ("FW;" + self._fw_name):
                self._fw_version = value
            elif key == "DEVICE;LIGHT":
                if self._has_light is None:
                    value = value.upper()
                    self._has_light = value in ("PRESENT", "PRESENT;COLOR")
            elif key == "NW;TOKEN":
                self._uuid = value.lower()
            elif key == "NAME;VALUE":
                self._name = value
            elif key == "GROUP;LIST":
                self._room_name = value
            elif key == "GROUP;ROOM;TYPE":
                self._room_type = int(value)
            elif key == "DEVICE;ID":
                self._mac, self._base_model = value.split(";")
                if self.model in ["Haiku Fan", "Haiku Light"]:
                    self._has_sensor = True
                elif self.model == "Haiku L Fan":
                    # determined by "DEVICE;OPTION;SENSORS" below
                    self._has_sensor = None
                else:
                    self._has_sensor = False
            elif key == "DEVICE;OPTION;SENSORS":
                if self._has_sensor is None:
                    value = value.upper()
                    self._has_sensor = value == "PRESENT"
        return "", should_callback

    def _send_update(self):
        """Sends update commands to an already connected device."""
        self._send_command("DEVICE;ID;GET")
        self._send_command("SNSROCC;STATUS;GET")
        self._send_command("GETALL")
        _LOGGER.debug("%s: Status update", self.name)

    async def _updater(self):
        """Periodically update device parameters.

        This method is a coroutine.
        """
        while True:
            try:
                self._send_update()
                await asyncio.sleep(self.refresh_minutes * 60 + random.uniform(-10, 10))
            except asyncio.CancelledError:
                _LOGGER.debug("%s: Updater task cancelled", self.name)
                return
            except OSError:
                _LOGGER.debug(
                    "%s: Updater task error\n%s", self.name, traceback.format_exc()
                )
                await asyncio.sleep(self.refresh_minutes * 60 + random.uniform(-10, 10))
            except Exception:
                _LOGGER.error(
                    "%s: Unhandled updater task error\n%s",
                    self.name,
                    traceback.format_exc(),
                )
                await asyncio.sleep(10)
                raise

    async def _listener(self):
        """Task that listens for device status changes.

        Maintains an open socket to the device at all times.
        Lost connections are automatically reconnected.
        This method is a coroutine.
        """
        while True:
            try:
                if self._error_count > 10:
                    _LOGGER.error("%s: Listener task too many errors", self.name)
                    self._is_connected = False
                    self._updater_task.cancel()
                    if self._endpoint is not None:
                        self._endpoint.close()
                    self._endpoint = None
                    break
                if self._endpoint is None:
                    self._is_connected = False
                    self._endpoint = SensemeEndpoint()
                    try:
                        _LOGGER.debug("%s: Connecting", self.name)
                        await asyncio.get_running_loop().create_connection(
                            lambda: SensemeProtocol(self._name, self._endpoint),
                            self._address,
                            PORT,
                        )
                        _LOGGER.debug("%s: Creating Updater Task", self.name)
                        self._updater_task = asyncio.create_task(self._updater())
                        self._error_count = 0
                        self._is_connected = True
                        if self._connection_lost:
                            _LOGGER.warning(
                                "%s: Connection to address %s restored",
                                self.name,
                                self.address,
                            )
                            self._connection_lost = False
                        else:
                            _LOGGER.debug(
                                "%s: Connection to address %s successful",
                                self.name,
                                self.address,
                            )
                    except OSError:
                        _LOGGER.debug(
                            "%s: Connect failed, try again in a minute\n%s",
                            self.name,
                            traceback.format_exc(),
                        )
                        self._endpoint = None
                        await asyncio.sleep(60)
                        continue
                data = await self._endpoint.receive()
                if data is None:
                    # endpoint is closed, let task know it's time open another
                    _LOGGER.warning(
                        "%s: Connection to address %s lost", self.name, self.address
                    )
                    self._data = dict()
                    self._is_connected = False
                    self._first_update = False
                    self._connection_lost = True
                    self._execute_callbacks()  # tell callbacks we disconnected
                    self._endpoint = None
                    self._updater_task.cancel()
                    await asyncio.sleep(1)
                    continue
                # add previous partial data to new data and process
                self._leftover, should_callback = self._process_message(
                    self._leftover + data
                )
                if should_callback:
                    self._execute_callbacks()
            except asyncio.CancelledError:
                _LOGGER.debug("%s: Listener task cancelled", self.name)
                return
            except OSError:
                _LOGGER.debug(
                    "%s: Listener task\n%s", self.name, traceback.format_exc()
                )
                self._error_count += 1
                await asyncio.sleep(1)
            except Exception:
                _LOGGER.error(
                    "%s: Listener task error\n%s", self.name, traceback.format_exc()
                )
                raise
        _LOGGER.error("%s: Listener task ended", self.name)

    def start(self):
        """Start the async task to handle responses from the device."""
        if not self._is_running:
            self._listener_task = asyncio.create_task(self._listener())
            self._is_running = True
            _LOGGER.debug("%s: Started", self.name)

    def stop(self):
        """Signals thread to stop and returns immediately."""
        if self._is_running is True:
            if self._listener_task is not None:
                self._listener_task.cancel()
            if self._updater_task is not None:
                self._updater_task.cancel()
            self._is_running = False


class SensemeFan(SensemeDevice):
    """SensemeFan Class."""

    def __str__(self) -> str:
        """Return string representation of SensemeFan object."""
        string = f"Name: {self._name}"
        if self._room_name is not None:
            string += f", Room Name: {self._room_name}"
        string += f", Model: {self.model}"
        string += f", UUID: {self._uuid}"
        string += f", IP: {self._address}"
        string += f", MAC: {self._mac}"
        if self._fw_version is not None:
            string += f", FW Version: {self._fw_version}"
        if self._has_light is not None:
            string += f", Has Light: {self._has_light}"
        return string

    @property
    def is_fan(self) -> str:
        """Return True if this device is a fan."""
        return True

    @property
    def has_light(self) -> bool:
        """Return True if the fan has an installed light."""
        return self._has_light

    @property
    def fan_on(self) -> bool:
        """Return True when fan is on at any speed."""
        state = self._data.get("FAN;PWR", None)
        if state:
            return state == "ON"
        return None

    @fan_on.setter
    def fan_on(self, state: bool):
        """Set the fan power state."""
        state = "ON" if state else "OFF"
        self._send_command(f"FAN;PWR;{state}")

    @property
    def fan_speed(self) -> int:
        """Return the fan speed."""
        speed = self._data.get("FAN;SPD;ACTUAL", None)
        if speed:
            return int(speed)
        return None

    @fan_speed.setter
    def fan_speed(self, speed: int):
        """Set the fan speed."""
        if speed < 0:
            speed = 0
        elif speed > self.fan_speed_max:
            speed = self.fan_speed_max
        self._send_command(f"FAN;SPD;SET;{speed}")

    @property
    def fan_speed_min(self) -> int:
        """Return the fan speed minimum."""
        min_speed = self._data.get("FAN;SPD;MIN", None)
        if min_speed:
            return int(min_speed)
        return None

    @property
    def fan_speed_max(self) -> int:
        """Return the fan speed maximum."""
        max_speed = self._data.get("FAN;SPD;MAX", None)
        if max_speed:
            return int(max_speed)
        return None

    @property
    def fan_speed_limits(self) -> Tuple:
        """Return a tuple of the min/max fan speeds."""
        return self.fan_speed_min, self.fan_speed_max

    @property
    def fan_speed_limits_room(self) -> Tuple:
        """Return a tuple of the min/max fan speeds the room is configured to support.

        A room can limit the minimum/maximum fan speed while keeping the same number of
        speed settings. On the Haiku by BAF application this setting can be found by
        clicking the room info button. There must be at least one fan added to a room.
        """
        raw = self._data.get("FAN;BOOKENDS", None)
        if raw is None:
            return None
        values = raw.split(";")
        if len(values) != 2:
            return None
        min_speed = int(values[0])
        max_speed = int(values[1])
        return min_speed, max_speed

    @fan_speed_limits_room.setter
    def fan_speed_limits_room(self, speeds: Tuple):
        """Set a tuple of the min/max fan speeds the room is configured to support."""
        if speeds[0] >= speeds[1]:
            _LOGGER.error("Min speed cannot exceed max speed")
            return
        self._send_command(f"FAN;BOOKENDS;SET;{speeds[0]};{speeds[1]}")

    @property
    def fan_dir(self) -> str:
        """Return the fan direction."""
        return self._data.get("FAN;DIR", None)

    @fan_dir.setter
    def fan_dir(self, direction: str):
        """Set the fan direction."""
        if direction not in DIRECTIONS:
            raise ValueError(
                f"{direction} is not a valid direction. Must be one of {DIRECTIONS}"
            )
        self._send_command(f"FAN;DIR;SET;{direction}")

    @property
    def fan_whoosh_mode(self) -> bool:
        """Return True when fan whoosh mode is on."""
        state = self._data.get("FAN;WHOOSH;STATUS", None)
        if state:
            return state == "ON"
        return None

    @fan_whoosh_mode.setter
    def fan_whoosh_mode(self, state: bool):
        """Set the fan whoosh mode."""
        value = "ON" if state else "OFF"
        self._send_command(f"FAN;WHOOSH;{value}")

    @property
    def fan_autocomfort(self) -> str:
        """Get the auto comfort mode from the fan.

        'OFF' no automatic adjustment,
        'COOLING' increases fan speed as temp increases,
        'HEATING' means slow mixing of air while room is occupied and faster mix speeds
        while room is not occupied.
        'FOLLOWTSTAT' means change between 'COOLING' and 'HEATING based on thermostat.
        """
        return self._data.get("SMARTMODE;STATE", None)

    @fan_autocomfort.setter
    def fan_autocomfort(self, state: str):
        """Set the fan auto comfort mode.

        'OFF' no automatic adjustment.
        'COOLING' increases fan speed as temp increases.
        'HEATING' means slow mixing of air while room is occupied and faster mix speeds
        while room is not occupied.
        'FOLLOWTSTAT' means change between 'COOLING' and 'HEATING based on thermostat.
        """
        value = "ON" if state else "OFF"
        self._send_command(f"SMARTMODE;STATE;SET;{value}")

    @property
    def fan_smartmode(self) -> str:
        """Get the current smart mode from the fan.

        'OFF' no automatic adjustment.
        'COOLING' increases fan speed as temp increases.
        'HEATING' means slow mixing of air while room is occupied and faster mix speeds
                  while room is not occupied.
        """
        return self._data.get("SMARTMODE;ACTUAL", None)

    @property
    def fan_cooltemp(self) -> float:
        """Return the auto shutoff temperature for 'COOLING' smart mode in Celsius."""
        temp = int(self._data.get("LEARN;ZEROTEMP", None))
        if temp:
            return float(temp) / 100.0
        return None

    @fan_cooltemp.setter
    def fan_cooltemp(self, temp: float):
        """Set the auto shutoff temperature for 'COOLING' smart mode in Celsius."""
        # force temperature into range
        if temp < 10:
            temp = 10
        elif temp > 31.5:
            temp = 31.5
        temp = int(round(temp * 100))
        self._send_command(f"LEARN;ZEROTEMP;SET;{temp}")

    @property
    def motion_fan_auto(self) -> bool:
        """Return True when fan is in automatic on with motion mode."""
        state = self._data.get("FAN;AUTO", None)
        if state:
            return state == "ON"
        return None

    @motion_fan_auto.setter
    def motion_fan_auto(self, state: bool):
        """Set the fan automatic on with motion mode."""
        state = "ON" if state else "OFF"
        self._send_command(f"FAN;AUTO;{state}")


class SensemeLight(SensemeDevice):
    """SensemeLight Class."""

    def __str__(self) -> str:
        """Return string representation of SensemeLight object."""
        string = f"Name: {self._name}"
        if self._room_name is not None:
            string += f", Room Name: {self._room_name}"
        string += f", Model: {self.model}"
        string += f", UUID: {self._uuid}"
        string += f", IP: {self._address}"
        string += f", MAC: {self._mac}"
        if self._fw_version is not None:
            string += f", FW Version: {self._fw_version}"
        return string

    @property
    def is_light(self) -> str:
        """Return True if the device is a standalone light."""
        return True

    @property
    def has_light(self) -> bool:
        """Return True if the fan has an installed light."""
        return True

    @property
    def light_color_temp(self) -> int:
        """Return the light color temperature."""
        color_temp = self._data.get("LIGHT;COLOR;TEMP;VALUE", None)
        if color_temp:
            return int(color_temp)
        return None

    @light_color_temp.setter
    def light_color_temp(self, color_temp: int):
        """Set the light color temperature."""
        if color_temp < self.light_color_temp_min:
            color_temp = self.light_color_temp_min
        if color_temp > self.light_color_temp_max:
            color_temp = self.light_color_temp_max
        color_temp = int(round(color_temp / 100.0)) * 100
        self._send_command(f"LIGHT;COLOR;TEMP;VALUE;SET;{color_temp}")

    @property
    def light_color_temp_min(self) -> int:
        """Return the light color temperature minimum (warmest light)."""
        min_color_temp = self._data.get("LIGHT;COLOR;TEMP;MIN", None)
        if min_color_temp:
            return int(min_color_temp)
        return None

    @property
    def light_color_temp_max(self) -> int:
        """Return the light color temperature maximum (coolest light)."""
        max_color_temp = self._data.get("LIGHT;COLOR;TEMP;MAX", None)
        if max_color_temp:
            return int(max_color_temp)
        return None



"""aiosenseme library."""

from .device import SensemeDevice, SensemeFan, SensemeLight
from .discovery import (
    SensemeDiscovery,
    async_get_device_by_device_info,
    async_get_device_by_ip_address,
    discover,
    discover_all,
    discover_any,
)
from .version import __version__

__all__ = [
    "SensemeDevice",
    "SensemeFan",
    "SensemeLight",
    "SensemeDiscovery",
    "discover_all",
    "discover_any",
    "discover",
    "async_get_device_by_ip_address",
    "async_get_device_by_device_info",
]

# flake8: noqa
